# Copyright 2014 Ross Delinger
# Use of this source code is governed by a BSD-style
# License that can be found in the LICENSE file.

from xml.etree import ElementTree
from sham.storage.pools import StoragePool
from sham.storage.volumes import Volume
from sham.network.interfaces import NetworkInterface


class VirtualMachine(object):
    """
    Wrapper around a libvirt Domain
    Provides a set of functions that make it easier to interact with
    domains.
    """
    def __init__(self, domain, vmm):
        self.vmm = vmm
        self.domain = domain
        self.name = domain.name()
        self.xml = ElementTree.fromstring(self.domain.XMLDesc(0))

    @property
    def domain_type(self):
        """
        Return what type of domain this this (kvm/lxc/etc)
        """
        return self.xml.attrib['type']

    @property
    def current_memory(self):
        """
        Return the current memory allocated to this VM
        """
        return self.xml.find('currentMemory').text

    @property
    def max_memory(self):
        """
        Return the maximum memory this VM can have
        """
        return self.xml.find('memory').text

    @property
    def num_cpus(self):
        """
        Return the number of Virtual CPUS allocated to this VM
        """
        return self.xml.find('vcpu').text

    def get_interfaces(self):
        """
        Return a list of sham.network.interfaces.NetworkInterface
        describing all the interfaces this VM has
        """
        interfaces = self.xml.find('devices').iter('interface')
        iobjs = []
        for interface in interfaces:
            _type = interface.attrib['type']
            mac = interface.find('mac').attrib['address']
            source = interface.find('source').attrib[_type]
            model = interface.find('model').attrib['type']
            iobjs.append(NetworkInterface(_type, mac, source, model))
        return iobjs

    def get_disks(self):
        """
        Return a list of all the Disks attached to this VM
        The disks are returned in a sham.storage.volumes.Volume
        object
        """
        disks = [disk for disk in self.xml.iter('disk')]
        disk_objs = []
        for disk in disks:
            source = disk.find('source')
            if source is None:
                continue
            path = source.attrib['file']
            diskobj = self.domain.connect().storageVolLookupByPath(path)
            disk_objs.append(diskobj)
        return [Volume(d, StoragePool(d.storagePoolLookupByVolume())) for d in disk_objs]

    def delete(self):
        """
        Delete this VM, and remove all its disks
        """
        disks = self.get_disks()
        self.domain.undefine()
        for disk in disks:
            disk.wipe()
            disk.delete()

    def start(self):
        """
        Start the current VM
        """
        self.domain.create()

    def shutdown(self, delete=False):
        """
        Shutdown this VM
        :param delete: Should we delete after shutting the VM down?
        :type delete: bool
        """
        disks = self.get_disks()
        self.domain.destroy()
        if delete:
            for disk in disks:
                disk.wipe()
                disk.delete()

    def is_running(self):
        """
        Check if this VM is running
        """
        info = self.domain.info()
        # info[0] is a state (see enum virDomainState)
        if info[0] == 1:
            return True
        return False

    def __repr__(self):
        return "VirtualMachine< {0} >".format(self.name)

    def to_dict(self):
        """
        Return the values contained in this object as a dict
        """
        return {'domain_type': self.domain_type,
                'max_memory': self.max_memory,
                'current_memory': self.current_memory,
                'num_cpus': self.num_cpus,
                'running': self.is_running(),
                'name': self.name,
                }



# Copyright 2014 Ross Delinger
# Use of this source code is governed by a BSD-style
# License that can be found in the LICENSE file.

from xml.etree import ElementTree


class NetworkInterface(object):
    def __init__(self, _type, mac_addr, source, model):
        self.xml_root = None
        self.type = _type
        self.mac = mac_addr
        self.source = source
        self.model = model

    def root(self):
        """
        TODO(rdelinger) rename this as to_xml or something similar
        """
        self.xml_root = ElementTree.Element('interface')
        self.xml_root.set('type', self.type)
        if self.mac is not None:
            mac = ElementTree.SubElement(self.xml_root, 'mac')
            mac.set('address', self.mac)
        _source = ElementTree.SubElement(self.xml_root, 'source')
        _source.set(self.type, self.source)
        _model = ElementTree.SubElement(self.xml_root, 'model')
        _model.set('type', self.model)
        return self.xml_root



# Copyright 2014 Ross Delinger
# Use of this source code is governed by a BSD-style
# License that can be found in the LICENSE file.
import libvirt
from libvirt import libvirtError
from xml.etree import ElementTree
from string import ascii_lowercase
from sham.storage.volumes import Volume
from sham.storage.pools import StoragePool
from sham.network.interfaces import NetworkInterface
from sham.machine import VirtualMachine


class RetryableHype(object):
    def __init__(self, uri):
        self.uri = uri
        self.hyp = libvirt.open(uri)

    def retry(self, func):
        def wraps(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except:
                print('Libvirt errored out, retrying')
                self.hyp = libvirt.open(self.uri)
                return getattr(self.hyp, func.__name__)(*args, **kwargs)
        return wraps

    def __getattr__(self, name):
        return self.retry(getattr(self.hyp, name))


class VMManager(object):
    def __init__(self, uri):
        self.hyper = RetryableHype(uri)

    def get_storage_pools(self):
        pools = self.hyper.listAllStoragePools(0)
        return [StoragePool(pool) for pool in pools]

    def memory_free(self):
        """
        Returns in GB
        """
        return self.hyper.getFreeMemory() / 1e9

    def get_vms(self):
        domains = self.hyper.listAllDomains(0)
        return [VirtualMachine(d, self) for d in domains]

    def find_vm(self, name):
        """
        Try and find a VM by name
        :param name: Name of the VM
        :type name: str
        """
        try:
            domain = self.hyper.lookupByName(name)
            VM = VirtualMachine(domain, self)
        except libvirtError:
            VM = None
        return VM

    def fast_clone(self, VM, clone_name, mem=None):
        """
        Create a 'fast' clone of a VM. This means we make
        a snapshot of the disk and copy some of the settings
        and then create a new VM based on the snapshot and settings
        The VM is transient so when it is shutdown it deletes itself
        :param VM: The VM to base this clone on
        :type VM: sham.machine.VirtualMachine

        :param clone_name: The name for this clone
        :type clone_name: str
        """
        disks = VM.get_disks()
        ints = VM.get_interfaces()
        count = 0
        new_disks = []
        for disk in disks:
            pool = disk.pool
            new_disk_name = '{0}-disk{1}'.format(clone_name, count)
            count += 1
            new_disk = pool.create_backed_vol(new_disk_name, disk)
            new_disks.append(new_disk)

        for inter in ints:
            inter.mac = None
            # if the mac is set to None we don't include it in the xml
            # and libvirt will autogen one for us
        return self.create_vm(
            VM.domain_type,
            clone_name,
            VM.num_cpus,
            mem or VM.current_memory,
            mem or VM.max_memory,
            new_disks,
            ints)

    def create_vm(self, domain_type, name, cpus, mem, memmax, disks, nets, memunit='KiB', arch='x86_64'):
        domain = ElementTree.Element('domain')
        domain.set('type', domain_type)

        dname = ElementTree.SubElement(domain, 'name')
        dname.text = name

        memory = ElementTree.SubElement(domain, 'memory')
        memory.set('unit', memunit)
        memory.text = memmax

        curmem = ElementTree.SubElement(domain, 'currentMemory')
        curmem.set('unit', memunit)
        curmem.text = mem

        vcpus = ElementTree.SubElement(domain, 'vpu')
        vcpus.text = cpus

        os = ElementTree.SubElement(domain, 'os')
        os_type = ElementTree.SubElement(os, 'type')
        os_type.set('arch', arch)
        os_type.text = 'hvm'  # TODO(rdelinger) Make this user setable
        os_boot = ElementTree.SubElement(os, 'boot')
        os_boot.set('dev', 'hd')

        devices = ElementTree.SubElement(domain, 'devices')
        devices_emu = ElementTree.SubElement(devices, 'emulator')
        devices_emu.text = '/usr/bin/qemu-system-x86_64'  # TODO(rdelinger) make this use setable

        target_drives = iter(['vd{0}'.format(letter) for letter in ascii_lowercase])
        for disk in disks:
            disk_elem = ElementTree.SubElement(devices, 'disk')
            disk_elem.set('type', 'file')
            disk_elem.set('device', 'disk')

            driver = ElementTree.SubElement(disk_elem, 'driver')
            driver.set('name', 'qemu')  # TODO(rdelinger) auto load this somehow?
            driver.set('type', disk.format)

            source = ElementTree.SubElement(disk_elem, 'source')
            source.set('file', disk.path)

            target = ElementTree.SubElement(disk_elem, 'target')
            target.set('dev', target_drives.next())
            target.set('bus', 'virtio')  # TODO(rdelinger) auto load this

        for net in nets:  # These are network interfaces not libvirt Networks
            devices.append(net.root())

        video = ElementTree.SubElement(devices, 'video')
        vmodel = ElementTree.SubElement(video, 'model')
        vmodel.set('type', 'cirrus')
        vmodel.set('vram', '9216')
        vmodel.set('heads', '1')

        graphics = ElementTree.SubElement(devices, 'graphics')
        graphics.set('type', 'vnc')
        graphics.set('port', '-1')
        graphics.set('autoport', 'yes')

        self.hyper.createXML(ElementTree.tostring(domain), 0)
        return self.find_vm(name)



# Copyright 2014 Ross Delinger
# Use of this source code is governed by a BSD-style
# License that can be found in the LICENSE file.

from xml.etree import ElementTree
from libvirt import libvirtError
from sham.storage.volumes import Volume


class StoragePool(object):
    def __init__(self, virsp):
        self.virsp = virsp

    def get_volumes(self):
        """
        Return a list of all Volumes in this Storage Pool
        """
        vols = [self.find_volume(name) for name in self.virsp.listVolumes()]
        return vols

    def create_backed_vol(self, name, backer, _format='qcow2'):
        """
        TODO(rdelinger) think about changing _format
        This is a pretty specialized function.
        It takes an existing volume, and creates a new volume
        that is backed by the existing volume
        Sadly there is no easy way to do this in libvirt, the
        best way I've found is to just create some xml and use the createXML
        function
        """
        vol_xml = ElementTree.Element('volume')
        vol_name = ElementTree.SubElement(vol_xml, 'name')
        name = '{0}.{1}'.format(name, _format)
        vol_name.text = name

        target = ElementTree.SubElement(vol_xml, 'target')
        target_format = ElementTree.SubElement(target, 'format')
        target_format.set('type', _format)

        vol_cap = ElementTree.SubElement(vol_xml, 'capacity')
        vol_cap.set('unit', 'bytes')  # @TODO(rdelinger) this should be dynamic
        vol_cap.text = backer.capacity

        backing_store = ElementTree.SubElement(vol_xml, 'backingStore')
        bs_path = ElementTree.SubElement(backing_store, 'path')
        bs_path.text = backer.path

        bs_format = ElementTree.SubElement(backing_store, 'format')
        bs_format.set('type', backer.format)

        XMLString = ElementTree.tostring(vol_xml)
        self.virsp.createXML(XMLString, 0)

        return self.find_volume(name)

    def find_volume(self, name):
        """
        Find a storage volume by its name
        :param name: The name of the volume
        :type name: str
        """
        try:
            return Volume(self.virsp.storageVolLookupByName(name), self)
        except libvirtError:
            return None

    def __repr__(self):
        name = self.virsp.name()
        numvols = self.virsp.numOfVolumes()
        return 'StoragePool<Name: {0}, Volumes: {1}>'.format(name, numvols)

